import { writable, readable, derived, get, type Subscriber } from 'svelte/store';
import {
  FedwaveChat,
  type Message,
  type OutgoingMessage,
  type Credentials
} from 'fedwave-chat-client';
import { globalMode } from './settings';
import * as jose from 'jose';

export const chat = new FedwaveChat();

export type TaggedMessage = {
  m: Message;
  id: string;
  isHeader: boolean;
};

export function send(m: string | OutgoingMessage) {
  chat.sendMessage(m);
}

let setConnected: Subscriber<boolean> | undefined;
export const connected = readable<boolean>(false, (setf) => {
  setConnected = setf;
  return () => {
    setConnected = () => {};
  };
});

function getCredentialsFromLocalStorage() {
  try {
    const token: Credentials | null = JSON.parse(localStorage.getItem('chatToken') || '""') || null;
    if (token !== null) {
      return token;
    }
  } catch (e) {
    console.error(e);
  }
}

let setCredentials: Subscriber<Credentials> | undefined;
export const credentials = readable<Credentials>(getCredentialsFromLocalStorage(), (setf) => {
  setCredentials = setf;
  return () => {
    setCredentials = () => {};
  };
});

export const updateCredentials = (creds: Credentials) => {
  setCredentials?.(creds);

  if (get(connected)) {
    chat.disconnect();
    chat.connect(get(room), creds);
  }
};

chat.onUpdateCredentials = async function (creds) {
  setCredentials?.(creds);
};

export const decodedToken = derived(credentials, (creds) => {
  const token = typeof creds == 'string' ? creds : creds?.jwt || '';
  let r;
  try {
    r = jose.decodeJwt(token).sub as any;
  } catch {}
  return r;
});

export const username = derived(decodedToken, (token) => {
  return token?.username;
});

export const trollRegex = /^troll:[\w]{4}$/i;
export const isTroll = derived(username, (usr) => {
  return trollRegex.test(usr);
});

export const color = derived(decodedToken, (token) => {
  return token?.color;
});

export const num = derived(decodedToken, (token) => {
  return token?.num;
});

export const room = writable<string>(chat.room);
room.subscribe((room) => {
  chat.room = room;
});

globalMode.subscribe((v) => {
  chat.global = v;
});

export const shouldTweet = writable(false);

// Essentially a shared_ptr idiom. When nobody is subscribed to this, the chat
// d/c's.
export const chat_lock = readable(null, () => {
  const release_connected = connected.subscribe(() => {});

  const creds = get(credentials);

  chat.connect(chat.room, creds).then(() => setConnected?.(true));

  return () => {
    chat.disconnect();
    setConnected?.(false);

    release_connected();
  };
});

export const messageLimit = writable<number>(200);

export const usernameRegex = derived(username, (user) => {
  return new RegExp(`(@${user})\\b`, 'gi');
});

export const mentions = writable<TaggedMessage[]>([]);

export const messages = readable<TaggedMessage[]>([], (set) => {
  chat.rcvMessageBulk = function (ms) {
    const usrRegex = get(usernameRegex);
    const old = get(messages);
    let prev = old[old.length - 1]?.m ?? { username: null, timestamp: 0 };

    let containsUsername = false;

    for (let i = 0; i < ms.length; ++i) {
      ms[i].message = ms[i].message.replaceAll(usrRegex, `<span class="mention">$1</span>`);
      const tagged: TaggedMessage = {
        m: ms[i],
        id: ms[i].username + ms[i].timestamp,
        isHeader:
          prev.username != ms[i].username || Math.abs(prev.timestamp - ms[i].timestamp) > 60000
      };

      if (usrRegex.test(ms[i].message)) {
        mentions.update((m) => m.concat({ ...tagged, isHeader: true }));
        containsUsername = true;
      }

      old.push(tagged);

      prev = ms[i];
    }

    const limit = get(messageLimit);
    if (old.length > limit) {
      old.splice(0, old.length - limit);
      old[0].isHeader = true;
    }

    shouldTweet.set(containsUsername);
    set(old);
  };

  return () => {
    chat.rcvMessageBulk = function () {
      return;
    };
  };
});

export const localMessages = derived([messages, room, globalMode], ([ms, room, global]) => {
  if (global) return ms;
  return ms.filter((m) => m.m.channel.toLowerCase() === room.toLowerCase());
});
